// (c) by Stefan Roettger, licensed under LGPL 2.1

#include <vector>

#include "minicrypt.h"

uint32_t minicrypt::Q[4096] = {0};
uint32_t minicrypt::q = 0;

uint32_t minicrypt::p = 0;
uint32_t minicrypt::c = 0;

// Adler-32 hash function
uint32_t minicrypt::hash(const std::string &buffer)
   {
   static const uint32_t prime = 65521;

   const uint8_t *buf = (uint8_t*)buffer.c_str();
   size_t n = buffer.size();

   uint32_t s1 = 1;
   uint32_t s2 = 0;

   for (size_t i=0; i<n; i++)
      {
      s1 = (s1 + buf[i]) % prime;
      s2 = (s2 + s1) % prime;
      }

   return((s2 << 16) | s1);
   }

// CMWC seed function
void minicrypt::seed(const std::string &key)
   {
   uint32_t h = hash(key);

   for (uint32_t i=0; i<4096; i++)
      {
      Q[i] = h;
      h = 271 * h + 13;
      }

   q = 4095;

   p = 18782; // prime
   c = 123;   // 0 <= c < p
   }

// CMWC random function
uint32_t minicrypt::random()
   {
   uint64_t t;
   uint32_t x;

   q = (q + 1) & 4095;
   t = ((uint64_t)p * Q[q]) + c;
   c = t >> 32;
   x = t + c;

   if (x < c) { ++x; ++c; }

   Q[q] = 0xfffffffe - x;

   return Q[q];
}

// encrypt with key
void minicrypt::encrypt(const std::string &key)
   {
   if (size() > 0)
      {
      seed(key);
      xor_buffer();
      seed(reverse(key));
      permute_buffer();
      }
   }

// decrypt with key
void minicrypt::decrypt(const std::string &key)
   {
   if (size() > 0)
      {
      seed(reverse(key));
      unpermute_buffer();
      seed(key);
      xor_buffer();
      }
   }

// permute buffer (Fisher-Yates shuffle)
void minicrypt::permute_buffer()
   {
   static const size_t c = 10000;

   size_t i = 0, n = size();
   size_t d = (random() % c) + 1;

   while (i+d < n)
   {
      permute_buffer(i,d);
      i += d;
      d = (random() % c) + 1;
   }

   permute_buffer(i,n-i);
   }

// permute buffer range (Fisher-Yates shuffle)
void minicrypt::permute_buffer(size_t s,size_t n)
   {
   if (n > 1)
      for (size_t i=n-1; i>0; i--)
         {
         size_t k = random() % i;

         char v = at(s+i);
         at(s+i) = at(s+k);
         at(s+k) = v;
         }
   }

// unpermute buffer (inverse Fisher-Yates shuffle)
void minicrypt::unpermute_buffer()
   {
   static const size_t c = 10000;

   size_t i = 0, n = size();
   size_t d = (random() % c) + 1;

   while (i+d < n)
   {
      unpermute_buffer(i,d);
      i += d;
      d = (random() % c) + 1;
   }

   unpermute_buffer(i,n-i);
   }

// unpermute buffer (inverse Fisher-Yates shuffle)
void minicrypt::unpermute_buffer(size_t s,size_t n)
   {
   if (n > 1)
      if (n < 65536)
         {
         std::vector<unsigned short int> b;

         for (unsigned int i=n-1; i>0; i--)
            {
            unsigned int k = random() % i;
            b.push_back(k);
            }

         for (unsigned int i=1; i<n; i++)
            {
            unsigned int k = b.back();
            b.pop_back();

            char v = at(s+i);
            at(s+i) = at(s+k);
            at(s+k) = v;
            }
         }
      else
         {
         std::vector<size_t> b;

         for (size_t i=n-1; i>0; i--)
            {
            size_t k = random() % i;
            b.push_back(k);
            }

         for (size_t i=1; i<n; i++)
            {
            size_t k = b.back();
            b.pop_back();

            char v = at(s+i);
            at(s+i) = at(s+k);
            at(s+k) = v;
            }
         }
   }

// xor buffer with random values
void minicrypt::xor_buffer()
   {
   size_t n = size();

   for (size_t i=0; i<n; i++)
      at(i) ^= (random() >> 24);
   }

// reverse string
std::string minicrypt::reverse(const std::string s)
   {
   std::string r;

   std::string::const_reverse_iterator it = s.rbegin();

   while (it != s.rbegin())
      r.push_back(*it++);

   return(r);
   }
