// (c) by Stefan Roettger, licensed under LGPL 2.1

#ifndef MINIMEAS_H
#define MINIMEAS_H

#include "minicoord.h"
#include "ministring.h"

//! geo-referenced measurements
class minimeas: public minicoord
   {
   public:

   //! default constructor
   minimeas()
      : minicoord(),
        accuracy(NAN), velocity(NAN), heading(NAN), inclination(NAN),
        start(FALSE),
        power(NAN), frequency(NAN), heartbeat(NAN),
        description(NULL), metadata(NULL),
        flags(0)
      {}

   //! constructor
   minimeas(const minicoord &c,
            float a=NAN,float v=NAN,float h=NAN,float i=NAN,
            BOOLINT s=FALSE,
            float p=NAN,float f=NAN,float b=NAN)
      : minicoord(c),
        accuracy(a), velocity(v), heading(h), inclination(i),
        start(s),
        power(p), frequency(f), heartbeat(b),
        description(NULL), metadata(NULL)
      {
      if (heading<-180.0) heading+=360.0;
      else if (heading>180.0) heading-=360.0;
      }

   //! copy constructor
   minimeas(const minimeas &m)
      : minicoord((minicoord)m),
        accuracy(m.accuracy), velocity(m.velocity), heading(m.heading), inclination(m.inclination),
        start(m.start),
        power(m.power), frequency(m.frequency), heartbeat(m.heartbeat),
        flags(m.flags)
      {
      if (m.description) description=new ministring(*m.description);
      else description=NULL;

      if (m.metadata) metadata=new ministring(*m.metadata);
      else metadata=NULL;
      }

   //! destructor
   ~minimeas()
      {
      if (description)
         delete description;

      if (metadata)
         delete metadata;
      }

   //! assignment operator
   minimeas& operator = (const minimeas &m)
      {
      *(minicoord *)this=(minicoord)m;

      accuracy=m.accuracy;
      velocity=m.velocity;
      heading=m.heading;
      inclination=m.inclination;

      start=m.start;

      power=m.power;
      frequency=m.frequency;
      heartbeat=m.heartbeat;

      if (m.description) description=new ministring(*m.description);
      else description=NULL;

      if (m.metadata) metadata=new ministring(*m.metadata);
      else metadata=NULL;

      flags=m.flags;

      return(*this);
      }

   //! set description
   void set_description(const ministring &desc)
      {
      if (description)
         {
         delete description;
         description=NULL;
         }

      if (!desc.empty())
         {
         description=new ministring(desc);

         description->remove_leading_white_space();
         description->remove_trailing_white_space();
         }
      }

   //! get description
   ministring get_description() const
      {
      if (description)
         return(*description);

      return("");
      }

   //! set metadata
   void set_metadata(const ministring &meta)
      {
      if (metadata)
         {
         delete metadata;
         metadata=NULL;
         }

      if (!meta.empty())
         {
         metadata=new ministring(meta);

         metadata->remove_leading_white_space();
         metadata->remove_trailing_white_space();
         }
      }

   //! get metadata
   ministring get_metadata() const
      {
      if (metadata)
         return(*metadata);

      return("");
      }

   //! set flag
   void set_flag(int n,BOOLINT on=TRUE)
      {
      if (on) flags|=(1<<n);
      else flags&=~(1<<n);
      }

   //! get flag
   BOOLINT get_flag(int n) const
      {return((flags&(1<<n))!=0);}

   //! set flags
   void set_flags(unsigned int v)
      {flags=v;}

   //! set flags
   void set_flags(const char *meta)
      {set_flags(compute_flags(meta));}

   //! set flags
   void set_flags(const ministring &meta)
      {set_flags(compute_flags(meta));}

   //! compute flags
   static unsigned int compute_flags(const char *meta)
      {
      unsigned int f=0;
      while (*meta!='\0')
         {
         char c=*meta;
         if (c>='a' && c<='z') f|=(1<<(c-'a'));
         else if (c>='A' && c<='Z') f|=(1<<(c-'A'));
         meta++;
         }
      return(f);
      }

   //! compute flags
   static unsigned int compute_flags(const ministring &meta)
      {return(compute_flags(meta.cc_str()));}

   //! check flags
   BOOLINT check_flags(char c) const
      {
      if (c>='a' && c<='z') return(get_flag(c-'a'));
      else if (c>='A' && c<='Z') return(get_flag(c-'A'));
      return(FALSE);
      }

   //! check flags
   BOOLINT check_flags() const
      {return(flags!=0);}

   //! serialization
   ministring to_string() const;

   //! deserialization
   void from_string(ministring &info);

   float accuracy;
   float velocity;
   float heading;
   float inclination;

   BOOLINT start;

   float power;
   float frequency;
   float heartbeat;

   protected:

   ministring *description;
   ministring *metadata;

   unsigned int flags;
   };

// arithmetic inline operators
inline minimeas operator + (const minimeas &a,const minimeas &b);
inline minimeas operator - (const minimeas &a,const minimeas &b);
inline minimeas operator * (const double a,const minimeas &b);
inline minimeas operator * (const minimeas &a,const double b);
inline minimeas operator / (const minimeas &a,const double b);
inline int operator == (const minimeas &a,const minimeas &b);
inline int operator != (const minimeas &a,const minicoord &b);
inline int operator < (const minimeas &a,const minimeas &b);

// output stream operator
inline std::ostream& operator << (std::ostream &out,const minimeas &m);

// implementation of inline operators:

inline minimeas operator + (const minimeas &a,const minimeas &b)
   {
   return(minimeas(minicoord(a.vec+b.vec,a.type,a.crs_zone,a.crs_datum),
                   a.accuracy+b.accuracy,a.velocity+b.velocity,a.heading+b.heading,a.inclination+b.inclination,
                   a.start));
   }

inline minimeas operator - (const minimeas &a,const minimeas &b)
   {
   return(minimeas(minicoord(a.vec-b.vec,a.type,a.crs_zone,a.crs_datum),
                   a.accuracy+b.accuracy,a.velocity+b.velocity,a.heading-b.heading,a.inclination-b.inclination,
                   a.start));
   }

inline minimeas operator * (const double a,const minimeas &b)
   {
   return(minimeas(minicoord(a*b.vec,b.type,b.crs_zone,b.crs_datum),
                   a*b.accuracy,a*b.velocity,b.heading,b.inclination,
                   b.start));
   }

inline minimeas operator * (const minimeas &a,const double b)
   {
   return(minimeas(minicoord(a.vec*b,a.type,a.crs_zone,a.crs_datum),
                   a.accuracy*b,a.velocity*b,a.heading,a.inclination,
                   a.start));
   }

inline minimeas operator / (const minimeas &a,const double b)
   {
   return(minimeas(minicoord(a.vec/b,a.type,a.crs_zone,a.crs_datum),
                   a.accuracy/b,a.velocity/b,a.heading,a.inclination,
                   a.start));
   }

inline int operator == (const minimeas &a,const minimeas &b)
   {return((minicoord)a==(minicoord)b);}

inline int operator != (const minimeas &a,const minimeas &b)
   {return((minicoord)a!=(minicoord)b);}

inline int operator < (const minimeas &a,const minimeas &b)
   {return(a.vec.w<b.vec.w);}

inline std::ostream& operator << (std::ostream &out,const minimeas &m)
   {return(out << m.to_string().c_str());}

#endif
