// (c) by Stefan Roettger, licensed under LGPL 2.1

#ifndef MINIWLS_H
#define MINIWLS_H

#include "vector"

#include "glslmath.h"

typedef struct {
   double t,m,w;
} WLS_Point;

typedef std::vector<WLS_Point> WLS_Series;

//! weighted least squares fit
class miniWLS: public WLS_Series
   {
   public:

   //! default constructor
   miniWLS() : WLS_Series()
      {c=vec3(0,0,0);}

   //! destructor
   ~miniWLS() {}

   //! fit curve (linear polynomial)
   void linear_fit();

   //! fit curve (quadratic polynomial)
   void quadratic_fit();

   //! evaluate fitted curve
   double eval(double t)
      {return(c.x+(c.y+c.z*t)*t);}

   //! get curve coefficients
   vec3 coefficients()
      {return(c);}

   protected:

   vec3 c;
   };

//! weighted least squares fit of angular quantities (degrees)
class miniWLSA: public miniWLS
   {
   public:

   //! fit curve (linear polynomial)
   void linear_fit()
      {
      limit();
      miniWLS::linear_fit();
      }

   //! fit curve (quadratic polynomial)
   void quadratic_fit()
      {
      limit();
      miniWLS::quadratic_fit();
      }

   //! evaluate fitted curve
   double eval(double t)
      {
      double a=miniWLS::eval(t);
      while (a>180) a-=360;
      while (a<-180) a+=360;
      return(a);
      }

   protected:

   void limit()
      {
      for (unsigned int i=1; i<size(); i++)
         {
         while (at(i).m-at(i-1).m>180) at(i).m-=360;
         while (at(i-1).m-at(i).m>180) at(i).m+=360;
         }
      }

   };

typedef struct {
   vec3 m;
   double t,w;
} WLS_Point3D;

typedef std::vector<WLS_Point3D> WLS_Series3D;

//! weighted 3D least squares fit
class miniWLS3D: public WLS_Series3D
   {
   public:

   //! default constructor
   miniWLS3D() : WLS_Series3D()
      {a=b=vec3(0,0,0);}

   //! destructor
   ~miniWLS3D() {}

   //! fit curve (linear parametric curve)
   void linear_fit();

   //! fit curve (parabolic parametric curve)
   void parabolic_fit();

   //! evaluate fitted curve
   vec3 eval(double t)
      {return(a+(b+c*t)*t);}

   //! get curve parameters
   void coefficients(vec3 &c1,vec3 &c2)
      {
      c1=a;
      c2=b;
      }

   //! get curve parameters
   void coefficients(vec3 &c1,vec3 &c2,vec3 &c3)
      {
      c1=a;
      c2=b;
      c3=c;
      }

   protected:

   vec3 a,b,c;
   };

#endif
